--[[
supplyLinesCore.lua
v1.5
by Knighttime

This file contains predefined core functionality for the Supply Lines module, and is not intended to be edited by the scenario designer.
]]

local mapWidth, mapHeight, mapQuantity = civ.getAtlasDimensions()
local supplyLineFound = nil
local validSupplyZoneTiles = { }
local validSupplyLine = { }
local supplyLinesCore = { }

-- ================================================
--  HELPER FUNCTIONS 
-- ================================================
-- Two helper functions to identify whether or not two tribes are allies, or at war:
supplyLinesCore.isAlly = function (tribe1, tribe2)
	local tribesAreAllies = false
	if tribe1.id ~= 0 and tribe2.id ~= 0 and tribe1.treaties[tribe2] & 0x0008 > 0 and tribe2.treaties[tribe1] & 0x0008 > 0 then
		tribesAreAllies = true
	end
	return tribesAreAllies
end
supplyLinesCore.isEnemy = function (tribe1, tribe2)
	local tribesAreEnemies = false
	if tribe1.id == 0 or tribe2.id == 0 or tribe1.treaties[tribe2] & 0x2000 > 0 or tribe2.treaties[tribe1] & 0x2000 > 0 then
		tribesAreEnemies = true
	end
	return tribesAreEnemies
end

-- Five helper functions to identify whether a tile contains a city and potentially the relationship of the city's owner to your tribe:
supplyLinesCore.containsCity = function (tile)
	local containsCity = false
	if tile.city ~= nil then
		containsCity = true
	end
	return containsCity
end
supplyLinesCore.containsAlliedTribeCity = function (tile, forTribe)
	local containsCity = false
	if tile.city ~= nil and tile.city.owner ~= forTribe and isAlly(tile.city.owner, forTribe) == true then
		containsCity = true
	end
	return containsCity
end
supplyLinesCore.containsEnemyTribeCity = function (tile, forTribe)
	local containsCity = false
	if tile.city ~= nil and tile.city.owner ~= forTribe and isEnemy(tile.city.owner, forTribe) == true then
		containsCity = true
	end
	return containsCity
end
supplyLinesCore.containsOtherTribeCity = function (tile, forTribe)
	local containsCity = false
	if tile.city ~= nil and tile.city.owner ~= forTribe then
		containsCity = true
	end
	return containsCity
end
supplyLinesCore.containsYourTribeCity = function (tile, forTribe)
	local containsCity = false
	if tile.city ~= nil and tile.city.owner == forTribe then
		containsCity = true
	end
	return containsCity
end

-- Five helper functions to identify whether a tile contains one or more units and potentially the relationship of the unit's owner to your tribe:
supplyLinesCore.containsUnit = function (tile)
	local hasUnit = false
	for unit in tile.units do
		hasUnit = true
		break
	end
	return hasUnit
end
supplyLinesCore.containsAlliedTribeUnit = function (tile, forTribe)
	local hasUnit = false
	for unit in tile.units do
		if unit.owner ~= forTribe then
			if isAlly(unit.owner, forTribe) == true then
				hasUnit = true
				break
			end
		end
	end
	return hasUnit
end
supplyLinesCore.containsEnemyTribeUnit = function (tile, forTribe)
	local hasUnit = false
	for unit in tile.units do
		if unit.owner ~= forTribe then
			if isEnemy(unit.owner, forTribe) == true then
				hasUnit = true
				break
			end
		end
	end
	return hasUnit
end
supplyLinesCore.containsOtherTribeUnit = function (tile, forTribe)
	local hasUnit = false
	for unit in tile.units do
		if unit.owner ~= forTribe then
			hasUnit = true
			break
		end
	end
	return hasUnit
end
supplyLinesCore.containsYourTribeUnit = function (tile, forTribe)
	local hasUnit = false
	for unit in tile.units do
		if unit.owner == forTribe then
			hasUnit = true
			break
		end
	end
	return hasUnit
end

-- Five helper functions to identify whether a tile contains a unit of a specific type and potentially the relationship of the unit's owner to your tribe:
supplyLinesCore.containsUnitType = function (tile, unittype)
	local hasUnittype = false
	for unit in tile.units do
		if unit.type == unittype then
			hasUnittype = true
			break
		end
	end
	return hasUnittype
end
supplyLinesCore.containsAlliedTribeUnitType = function (tile, unittype, forTribe)
	local hasUnittype = false
	for unit in tile.units do
		if unit.type == unittype and unit.owner ~= forTribe then
			if isAlly(unit.owner, forTribe) == true then
				hasUnittype = true
				break
			end
		end
	end
	return hasUnittype
end
supplyLinesCore.containsEnemyTribeUnitType = function (tile, unittype, forTribe)
	local hasUnittype = false
	for unit in tile.units do
		if unit.type == unittype and unit.owner ~= forTribe then
			if isEnemy(unit.owner, forTribe) == true then
				hasUnittype = true
				break
			end
		end
	end
	return hasUnittype
end
supplyLinesCore.containsOtherTribeUnitType = function (tile, unittype, forTribe)
	local hasUnittype = false
	for unit in tile.units do
		if unit.type == unittype and unit.owner ~= forTribe then
			hasUnittype = true
			break
		end
	end
	return hasUnittype
end
supplyLinesCore.containsYourTribeUnitType = function (tile, unittype, forTribe)
	local hasUnittype = false
	for unit in tile.units do
		if unit.type == unittype and unit.owner == forTribe then
			hasUnittype = true
			break
		end
	end
	return hasUnittype
end

-- =================================================
--  PRIMARY FUNCTIONS 
-- =================================================
supplyLinesCore.hasSupplyLine = function (supplyLines, gameObject, supplyDepotFunction, supplyDepotTileSet, supplyLinkFunction, minSupplyLineLength)
	local startTimestamp = os.clock()

	if supplyLines.SHOW_CONSOLE_MESSAGES then
		print("Detected call to supplyLinesCore.hasSupplyLine()")
	end

	-- NOTE: These are not global variables, rather they are defined as local at the top of this file, so they are scoped to all functions defined here:
	supplyLineFound = nil
	validSupplyZoneTiles = { }
	validSupplyLine = { }

	supplyLineFound = false
	local tilesEncountered = { }
	local tilesNotYetConsidered = { }
	local tileId = gameObject.x + (gameObject.y * mapWidth) + (gameObject.z * mapWidth * mapHeight)
	tilesEncountered[tileId] = 0
	table.insert(tilesNotYetConsidered, { tile = gameObject.location, path = { } })
	local countConsidered = 0

	while #tilesNotYetConsidered > 0 do
		local consideringTile = tilesNotYetConsidered[1]["tile"]
		local consideringTilePath = {table.unpack(tilesNotYetConsidered[1]["path"])}
		local consideringTileDistance = #consideringTilePath
		table.insert(consideringTilePath, consideringTile)
		countConsidered = countConsidered + 1
		table.remove(tilesNotYetConsidered, 1)
		local validSupplyDepot = false
		local description = ""
		if consideringTileDistance >= minSupplyLineLength then
			validSupplyDepot, description = supplyDepotFunction(consideringTile, consideringTileDistance, gameObject.owner, supplyDepotTileSet)
		end
		if validSupplyDepot == true then
			table.insert(validSupplyZoneTiles, consideringTile)
			supplyLineFound = true
			validSupplyLine = consideringTilePath
			if supplyLines.SHOW_CONSOLE_MESSAGES then
				print("  Found " .. description .. " at a distance of " .. tostring(#validSupplyLine - 1) .. " tiles")
			end
			break
		elseif supplyLinkFunction(consideringTile, gameObject.owner) == true then
			table.insert(validSupplyZoneTiles, consideringTile)
			if supplyLines.MAX_SUPPLY_LINE_LENGTH == nil or
			 consideringTileDistance < supplyLines.MAX_SUPPLY_LINE_LENGTH or
			 supplyLines.MAX_SUPPLY_LINE_LENGTH < 0 then

				local adjacentTiles = { }
				local xcoord = 0
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x - 2 else xcoord = (consideringTile.x - 2) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y, consideringTile.z))
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x - 1 else xcoord = (consideringTile.x - 1) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y - 1, consideringTile.z))
				xcoord = consideringTile.x
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y - 2, consideringTile.z))
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x + 1 else xcoord = (consideringTile.x + 1) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y - 1, consideringTile.z))
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x + 2 else xcoord = (consideringTile.x + 2) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y, consideringTile.z))
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x + 1 else xcoord = (consideringTile.x + 1) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y + 1, consideringTile.z))
				xcoord = consideringTile.x
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y + 2, consideringTile.z))
				if civ.game.rules.flatWorld == true then xcoord = consideringTile.x - 1 else xcoord = (consideringTile.x - 1) % mapWidth end
				table.insert(adjacentTiles, civ.getTile(xcoord, consideringTile.y + 1, consideringTile.z))

				local adjTileMustBeOceanOrCity = false
				if supplyLines.ONLY_COASTAL_CITIES_SUPPLY_ACROSS_OCEAN == true and consideringTile.baseTerrain.type == 10 then
					adjTileMustBeOceanOrCity = true
				end
				local adjTileDistance = consideringTileDistance + 1
				for _, adjTile in pairs(adjacentTiles) do
					local tileId = adjTile.x + (adjTile.y * mapWidth) + (adjTile.z * mapWidth * mapHeight)
					if tilesEncountered[tileId] == nil or tilesEncountered[tileId] > adjTileDistance then
						if adjTileMustBeOceanOrCity == false or adjTile.baseTerrain.type == 10 or adjTile.city ~= nil then
							tilesEncountered[tileId] = adjTileDistance
							table.insert(tilesNotYetConsidered, { tile = adjTile, path = consideringTilePath })
						end
--[[
						-- This would be helpful if all cities are supply depots and only cities are supply depots, since it would mark coastal
						-- 		terrain as within the supply zone. But since that's not guaranteed to be the case, it's probably misleading.
						if adjTileMustBeOceanOrCity == true and adjTile.baseTerrain.type ~= 10 and adjTile.city == nil then
							table.insert(validSupplyZoneTiles, adjTile)
						end
]]
					end
				end

			end
		end
	end

	if supplyLineFound == false then
		local message = "  No supply depot found"
		if supplyLines.MAX_SUPPLY_LINE_LENGTH >= 0 then
			message = message .. " within " .. supplyLines.MAX_SUPPLY_LINE_LENGTH .. " tiles"
		else
			message = message .. " on this map"
		end
		if supplyLines.SHOW_CONSOLE_MESSAGES then
			print(message)
		end
	end
	if supplyLines.SHOW_CONSOLE_MESSAGES then
		print("    Analyzed " .. countConsidered .. " tiles in " .. string.format("%.4f", os.clock() - startTimestamp) .. " seconds")
	end

	return supplyLineFound
end

supplyLinesCore.showSupplyLine = function (supplyLines)
	print("Detected call to showSupplyLine()")
	local supplyMarkersFound = 0
	-- Delete all existing markers:
	for unit in civ.iterateUnits() do
		if unit.type == supplyLines.SUPPLY_MARKER_UNITTYPE then
			local unitLocation = unit.location
			civ.deleteUnit(unit)
			civ.ui.redrawTile(unitLocation)
			supplyMarkersFound = supplyMarkersFound + 1
		end
	end
	print("Existing supply markers found: " .. supplyMarkersFound)
	print("Supply line found: " .. tostring(supplyLineFound))
	local supplyMarkersCreated = 0
	if supplyMarkersFound == 0 and supplyLineFound ~= nil then
		if supplyLines.SUPPLY_MARKER_UNITTYPE == nil then
			civ.ui.text("Unable to proceed. No valid definition of supplyLines.SUPPLY_MARKER_UNITTYPE within supplyLines.lua")
		else
			-- Create new markers:
			if supplyLineFound == true then
				for _, tile in pairs(validSupplyLine) do
					civ.createUnit(supplyLines.SUPPLY_MARKER_UNITTYPE, civ.getCurrentTribe(), tile)
					civ.ui.redrawTile(tile)
					supplyMarkersCreated = supplyMarkersCreated + 1
				end
			elseif supplyLineFound == false then
				if #validSupplyZoneTiles > supplyLines.SUPPLY_MARKER_MAX_QUANTITY then
					civ.ui.text("Unable to proceed. There are " .. #validSupplyZoneTiles .. " tiles within the supply zone, and markers are only supported for up to " .. supplyLines.SUPPLY_MARKER_MAX_QUANTITY .. " tiles.")
				else
					for _, tile in pairs(validSupplyZoneTiles) do
						civ.createUnit(supplyLines.SUPPLY_MARKER_UNITTYPE, civ.getCurrentTribe(), tile)
						civ.ui.redrawTile(tile)
						supplyMarkersCreated = supplyMarkersCreated + 1
					end
				end
			end
		end
	end
	print("New supply markers created: " .. supplyMarkersCreated)
end

print("supplyLinesCore.lua parsed successfully at " .. os.date("%c"))
print("    Lines found: " .. debug.getinfo(1).currentline + 8)

return supplyLinesCore
